import re


def find_all_index(pattern, string, flags=0):
    """find all matched index of string"""
    return [i.span() for i in re.finditer(pattern, string, flags=flags)]


def string_add(string: str, dx=1):
    # count_points = string.count('.')
    items = find_all_index(r"\.", string)
    number_list = [i for i in string.split('.')]
    number_str = "".join(number_list)
    number_len = len(number_str)
    number = int(number_str)
    number += dx
    new_number_str = f"{number:0>{number_len}d}"
    new_number_list = list(new_number_str)
    [new_number_list.insert(idx[0], ".") for idx in items]
    return "".join(new_number_list)



from .version_ops import yaml_load
from .file_ops import ppath


_version_config = yaml_load(ppath("version-config.yaml"))
__version__= _version_config['version']
print(f"{_version_config['name']} version is: {__version__}")



from .color_str import rgb_string, GREEN
from functools import wraps, partial
import logging
import time


def runtime(times=10):
    # if func is None:
    #     return partial(time_it, times=times)
    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time.time()
            for i in range(times):
                func(*args, **kwargs)
            end = time.time()
            average_cost_time = (end-start)/times
            time_str = f"{average_cost_time:.3f}"

            print(f"Run {rgb_string(str(times), GREEN)} times, "
                  f"the average time is {rgb_string(time_str, GREEN)} seconds.")
            return func(*args, **kwargs)
        return wrapper
    return decorate


def logged(level, name=None, message=None):
    """
    Add logging to a function. level is the logging
    level, name is the logger name, and message is the
    log message. If name and message aren't specified,
    they default to the function's module and name.
    """
    def decorate(func):
        logname = name if name else func.__module__
        log = logging.getLogger(logname)
        logmsg = message if message else func.__name__

        @wraps(func)
        def wrapper(*args, **kwargs):
            log.log(level, logmsg)
            return func(*args, **kwargs)
        return wrapper
    return decorate


if __name__ == "__main__":
    # Example use
    @logged(logging.DEBUG)
    def add(x, y):
        return x + y


    @logged(logging.CRITICAL, 'example')
    def spam():
        print('Spam!')


    @runtime(1)
    def function(n):
        s = -1
        for i in range(n):
            s += i
            time.sleep(0.1)
        return s

    print(function(10))


from collections import Counter
from functools import wraps
import numpy as np


def broadcast(func):  # It can be replaced by `np.vectorize`
    """
        example:
        @broadcast
        def f(x):
            # A function that can map only a single element
            if x==1 or x==0:
                return x
            else:
                return f(x-1)+f(x-2)

        >> f([2,4,10])
        >> array([1, 3, 832040], dtype=object)
    """

    @wraps(func)
    def wrap(*args, **kwargs):
        """
        Takes an arbitrary Python function and returns a NumPy ufunc
        Can be used, for example, to add broadcasting to a built-in Python function
        return: only one out,
        type:numpy object

        """
        nin, nout = len(args) + len(kwargs), 1
        return np.frompyfunc(func, nin, nout)(*args, **kwargs)

    return wrap


class _Dict_enhance(dict):
    """Enables the dictionary to be dot operated"""
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self


def dict_dotable(dic):
    """
    : input: a dictionary
    : output: an enhanced dictionary
    Example:
        enhance_dic = dict_dotable(dic)
    then, you can operate an enhanced dictionary like this:
        enhance_dic.key1.key2. ...
    """
    dic = _Dict_enhance(dic)
    for i in dic:
        if type(dic[i]) == dict:
            dic[i] = dict_dotable(dic[i])
    return dic


class Const:
    """
    define a constant like C language.
    `object.__setattr__(self, name, value)`
    this built-in function will called when assigning values to properties of the class

    `object.__dict__` holds all writable attributes in object,
    key as variable name and value as variable value.
    """

    def __setattr__(self, name, value):
        if hasattr(self, name):
            raise ValueError('Constant value can\'t be changed')
        else:
            self.__dict__[name] = value


def index_char(L=1000):
    """
    Get the index of all characters.
    use chr()
    """
    index_token = {}
    token = []
    for i in range(L):
        character = chr(i)
        index_token[i] = character
        token.append(character)
    token_index = dict(zip(token,
                           range(L)))  # token_index[idx] equal to ord(idx)
    return index_token, token_index


# Enables the dictionary to be dot operated
class _Dict_enhance(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self


def dict_dotable(dic):
    '''
    : input: a dictionary
    : output: an enhanced dictionary
    Example:
        enhance_dic = dict_dotable(dic)
    then, you can operate an enhanced dictionary like this:
        enhance_dic.key1.key2. ...
    '''
    dic = _Dict_enhance(dic)
    for i in dic:
        if type(dic[i]) == dict:
            dic[i] = dict_dotable(dic[i])
    return dic


def get_number_digits(number):
    res = number
    digit = 1
    if res >= 1:
        while res > 10:
            digit += 1
            # res, mod = np.divmod(res, 10)
            res //= 10
    else:
        while res < 1:
            digit -= 1
            res *= 10
    return digit


def get_digits(number_like):
    number_str = str(int(str(number_like)))
    return len(number_str)


def sort_count(lis):
    """
    返回lis的由大到小排序好的元素列表
    Example:
    l = np.array([2,2,2,2,5,5,3,9,9])
    sort_count(l) : [(2, 4), (5, 2), (9, 2), (3, 1)]
    # return [2, 5, 9,3], [4, 2, 2, 1]
    """
    a = Counter(lis)
    b = sorted(a.items(), key=lambda item: item[1], reverse=True)
    # idx, counts = [b[i][0] for i in range(len(b))], [b[i][1] for i in range(len(b))]
    return b


import os
from .file_ops import yaml_load, yaml_dump
from .string import string_add


class VersionControl:
    def __init__(self,
                 pkgname,
                 pkgdir,
                 version=None,
                 filename="version-config.yaml",
                 ):

        self.config = None
        self._pkgname = pkgname
        self._config_path = os.path.join(pkgdir, filename)
        if version is None:
            try:
                self.get_config()
            except:
                print(f"{filename} was not exist, created now.")
                self.gen_config('0.0.0')
        else:
            self.gen_config(version)

    def gen_config(self, version="0.0.0"):
        config = {"name": self._pkgname, "version": version}
        self.config = config
        yaml_dump(self._config_path, config)

    def get_config(self):
        config = yaml_load(self._config_path)
        self.config = config
        return config

    def set_version(self, version):
        self.config['version'] = version

    def save_config(self):
        yaml_dump(self._config_path, self.config)

    def update_version(self, version_step=1):
        self.config['version'] = string_add(self.config["version"], version_step)
        yaml_dump(self._config_path, self.config)

    def clean_config_file(self):
        os.remove(self._config_path)
